#!/usr/bin/env python

# Author                 Version      Date
# -----------------------------------------------
# S M Crawford (SAA0)    0.30         25 Oct 2008
#
# SLOTUTC is a tool to fix the UTC time in SALTICAM 
# slot mode imaging.  Because of a timing errror in 
# OS being used, the clock loses time between seconds 
# and the UTC time is inaccurate between updates of 
# the system clock (which is currently updated at the
# beginning of each second.
#
# As such this program reads in a set of slotmode data
# and calculates what the true exposure plus deadtime is.
# Once calculated, it will go through the images and 
# correct the UTC times to the correct value.  This process
# assumes that the UTC time for an image right after the 
# second turns is a fiducial time.  
#

import os,sys, time, re
from pyraf import iraf 
import saltprint, saltkey, saltio, salttime
import slottool

import pyfits
import Tkinter
from numpy import *
from pylab import *

# ----------------------------------------------------------
# core routine

def slotutcfix(images, update, ampperccd, inter, plotdata, logfile,verbose,status):



#set up the variables
    status = 0
    utc_list = []

# test the logfile

    logfile = saltio.logname(logfile)

# log the call 

    saltprint.line(logfile,verbose)
    message = 'SLOTUTCFIX -- '
    message += 'images='+images+' '
    yn = 'n'
    if (update): yn = 'y'
    message += 'update='+yn+' '
    yn = 'n'
    message += 'ampperccd=%i ' % ampperccd
    if (inter): yn = 'y'
    message += 'inter='+yn+' '
    yn = 'n'
    if (plotdata): yn = 'y'
    message += 'plotdata='+yn+' '
    message += 'logfile='+logfile+' '
    yn = 'n'
    if (verbose): yn = 'y'
    message += 'verbose='+yn+'\n'
    saltprint.log(logfile,message,verbose)

# start time

    saltprint.time('SLOTUTCFIX started at  ',logfile,verbose)

# is the input file specified?
   
    if (status == 0): status = saltio.filedefined('Input',images,logfile)

# if the input file is a list, does it exist?

    if (status == 0 and images[0] == '@'): 
        status = saltio.listexists('Input',images,logfile)

# parse list of input files

    if (status == 0):
        infiles, status = saltio.listparse('Raw image',images,'','','',logfile)

# check input files exist

    if (status == 0): status = saltio.filesexist(infiles,'','r',logfile)

# get time of first exposure and basic information about the observations

    if (status == 0):

	infile=infiles[0]
	struct, status = saltio.openfits(infile,logfile)

        # check to make sure slotmode data
        if (status ==0 ):
	   detmode, status = saltkey.get('DETMODE',struct[0], infile, logfile)
	   if detmode != 'Slot Mode':
              message='SLOTUTCFIX--ERROR: Data are not Slot Mode Observations'
              status=saltprint.err(logfile, message)

        #requested exposure time
        if (status == 0):
	   req_texp,status = saltkey.get('EXPTIME',struct[0],infile,logfile)
              

	# how many extensions?
    	if (status == 0): 
	   nextend,status = saltkey.get('NEXTEND',struct[0],infile,logfile)

	# how many amplifiers
    	if (status == 0): 
	   amplifiers,status = saltkey.get('NCCDS',struct[0],infile,logfile)
	   amplifiers = int(ampperccd*float(amplifiers))
	   if ampperccd>0:
	      nframes = nextend/amplifiers
	      nstep=amplifiers
	   else:
	      nframes = nextend
	      nstep=1

	#how many total frame and unique times
	if (status==0):
	   ntotal=nextend*len(infiles)
	   nunique=len(infiles)*nframes

    #Create arrays necessary for analysis
    if (status==0):
       id_arr=arange(nunique)
       utc_arr=zeros(nunique,dtype=float)

    #Read in each file and make a list of the UTC values
    if status==0:
      j=0
      for infile in infiles:

	if (status == 0 ): struct, status = saltio.openfits(infile,logfile)

	#Skip through the frames and read in the utc
        if (status == 0 ):
	   for i in range(1,len(struct), amplifiers):
	       utc_list.append(struct[i].header['UTC-OBS'])
	       utc_arr[j]=slottool.getobstime(struct[i], infile, logfile)
               j += 1
               

        # close FITS file		
	status = saltio.closefits(struct,logfile)

    print len(infiles), len(utc_list), len(utc_arr), nframes
    
    #set up the other important arrays 
    if (status == 0): 
        try:
           diff_arr=utc_arr.copy()
           diff_arr[1:]=diff_arr[1:]-utc_arr[:-1]
	   diff_arr[0]=-1
           dsec_arr=utc_arr-utc_arr.astype(int)
	except:
	   message='SLOTUTCFIX--ERROR:  Unable to create timing arrays'
           status=saltprint.err(logfile, message)
           


    #calculate the real exposure time
    if (status == 0):
	real_expt, med_expt, status= \
           calculate_realexptime(id_arr, utc_arr, dsec_arr, diff_arr, req_texp, logfile, utc_list)
	print med_expt, real_expt, req_texp
      



# end time

    if (status == 0):
	saltprint.time('SLOTUTCFIX completed at',logfile,verbose)
    else:
	saltprint.time('SLOTUTCFIX aborted at',logfile,verbose)

# -----------------------------------------------------------
# Calculate the actual exposure time

def calculate_realexptime(id_arr, utc_arr, dsec_arr, diff_arr, req_texp, logfile, utc_list):
    """Calculates the real exposure time.  This makes the following assumptions:
	1. That the measurement after the turn of the second is a fiducial
	2. That there is an integer number of frames between each fiducial exposure
        3. We then set up a metric which is Y=sum(i-int(i)) where i=dt/t_exp
        4. Then the minimum of Y is found between the requested exposure time and 
             the median time difference
        5. And the best exposure time is the time at that minimum
                
  
       returns median exposure time and real exposure time
    """
    status=0
    t_exp=0

    #calculate the median time
    try:
        t_wrong=median(diff_arr)
    except:
	message='SLOTUTCFIX--ERROR:  Unable to calculate median time difference'
        status=saltprint.err(logfile, message)

    
    #Compress the arrays to find those closest to the second mark
    if (status==0):
       mask=(dsec_arr<t_wrong)*(diff_arr>0)
       t=compress(mask,utc_arr)
       s=compress(mask,dsec_arr)
       id=compress(mask,id_arr)

    #Now set up the components in the equation
    if (status==0):
       try:
          dt=t[1:]-t[0]  #t1-t2 
       except Exception, e:
	  message='SLOTUTCFIX--ERROR:  Unable to set up necessary arrays because %s' % e
          status=saltprint.err(logfile, message)
	
    #Now find the one that minimizes the equation y=sum(i-int(i))
    if (status==0):
	t_real, status=find_real_time(dt, req_texp, t_max)

    return t_real, t_wrong, status

def find_times(dt):
   """Find the set of median times and how many times they appear

      return list, list
   """
   from sets import Set
 
   #find only the unique values
   sdt=Set(dt)

   mdt=zeros(len(sdt), dtype=float)
   cdt=zeros(len(sdt))

   i=0
   for t in sdt:
       mdt[i]=t
       mask=(dt==t)
       count=mask.sum()
       cdt[i]=count
       i += 1

   return mdt, cdt


def find_real_time(mdt, cmdt, t_min, t_max):
   """Find the real exposure+dead time
   
      returns float  
   """
   t_e=arange(0.9*t_min,1.2*t_max,0.0001)
   ysum=t_e*0.0
   for i in range(len(t_e)):
       ysum[i]=ntime_func(mdt, cmdt, t_e[i])

   print t_e[ysum.argmin()], t_min, t_max, ysum.min()

   t_r=0.1283
   y=0
   for j in range(len(mdt)):
        i=mdt[j]/t_r
        y += abs(i-round(i))
        #print j, i,i-round(i),  y, mdt[j] , cmdt[j]

   t_r=0.147
   y=0
   for j in range(len(mdt)):
        i=mdt[j]/t_r
        y += abs(i-round(i))
        #print j, i, i-round(i), y, mdt[j], cmdt[j]

   figure(figsize=(8,8), dpi=72)
   axes([0.1,0.1,0.8,0.8])
   plot(t_e, ysum,ls='-',marker='')
   show()


   return t_e[ysum.argmin()]
   
def ntime_func(mdt, cmdt, t_e):
    """Merit function to determine best time
       Weighted for the number of objects in each step
       return float
    """
    y=0
    ctot=cmdt.sum()
    for j in range(len(mdt)):
	i=mdt[j]/t_e
	y += abs(i-round(i))*cmdt[j]
    return y/ctot


# -----------------------------------------------------------
# main code 

parfile = iraf.osfn("slottools$slotutcfix.par") 
t = iraf.IrafTaskFactory(taskname="slotutcfix",value=parfile,function=slotutcfix,pkgname='slottools')
