#!/usr/bin/env python

# Author                 Version      Date
# -----------------------------------------------
# S M Crawford (SAAO)    0.3          10 Apr 2008

# slotmerge will combined images distributed over multiple 
# extensions into a single image.  It will be able to handle
# multiple images in a single FITS file and will also update the 
# header information with all of the relevant details.

from pyraf import iraf 

import saltprint, saltkey, saltio, saltstring, salttran
import os, time
import numpy, pyfits

# -----------------------------------------------------------
# core routine

def slotmerge(images,outimages,outpref,geomfile,clobber,logfile,verbose,status):

    status = 0

# test the logfile

    logfile = saltio.logname(logfile)

# log the call 

    saltprint.line(logfile,verbose)
    history = 'SLOTMERGE  -- '
    history += 'images='+images+' '
    history += 'outimages='+outimages+' '
    history1 = 'outpref='+outpref+' '
    history1 += 'geomfile='+geomfile+' '
    history1 += 'logfile='+logfile+' '
    yn = 'n'
    if (clobber): yn = 'y'
    history1 += 'clobber='+yn+' '
    yn = 'n'
    if (verbose): yn = 'y'
    history1 += 'verbose='+yn
    history = history + history1
    saltprint.log(logfile,history,verbose)

# start time

    saltprint.time('\nSLOTMERGE  -- started at',logfile,verbose)
    saltprint.log(logfile,' ',verbose)

# are the arguments defined
   
    if (status == 0): status = saltio.argdefined('images',images,logfile)
    if (status == 0): status = saltio.argdefined('geomfile',geomfile,logfile)
    if (status == 0): status = saltio.argdefined('logfile',logfile,logfile)

# if the input file is a list, does it exist?

    if (status == 0 and images[0] == '@'): 
        status = saltio.listexists('Input',images,logfile)

# parse list of input files

    if (status == 0):
        infiles, status = saltio.listparse('Raw image',images,'','','',logfile)

# check input files exist

    if (status == 0): status = saltio.filesexist(infiles,'','r',logfile)

# load output name list: @list, * and comma separated
   
    outimages = outimages.strip()
    outpref = outpref.strip()
    if (status == 0 and len(outpref) == 0 and len(outimages) == 0):
        message = '\nERROR -- SLOTMERGE : output file(s) not specified'
        status = saltprint.err(logfile,message)

# test output @filelist exists

    if (status == 0 and len(outimages) > 0 and outimages[0] == '@'): 
        status = saltio.listexists('Output',outimages,logfile)

# parse list of output files

    if (status == 0):
        outfiles, status = saltio.listparse('Output image',outimages,outpref,infiles,'',logfile)

# are input and output lists the same length?

    if (status == 0): status = saltio.comparelists(infiles,outfiles,'Input','output',logfile)

# do the output files already exist?

    if (status == 0 and not clobber): status = saltio.filesexist(outfiles,'','w',logfile)

# does CCD geometry definition file exist

    geomfilefile = geomfile.strip()
    if (status == 0): status = saltio.fileexists(geomfile,logfile)

# read geometry definition file

    gap = 0
    xshift = [0, 0]
    yshift = [0, 0]
    rotation = [0, 0]
    if (status == 0): 
         gap, xshift, yshift, rotation, status = saltio.readccdgeom(geomfile,logfile,status)
	 for ro in rotation:
	     if ro!=0:
		message='SLOTMERGE--WARNING:  SLOTMERGE currently ignores CCD rotation'
                saltprint.log(logfile,message,verbose)

# Begin processes each file 

    ifil = 0
    if status == 0:
     for infile in infiles:
	 pstatus=status

	 #determine the name for the output file
	 outfile = outfiles[ifil]
	 outpath = outfile.rstrip(os.path.basename(outfile))
	 if (len(outpath) == 0): outpath = '.'

	 #open each raw image
	 if (pstatus == 0):
	    struct, pstatus = saltio.openfits(infile,logfile)

	 # identify instrument

	 if (pstatus == 0): instrume,keyprep,keygain,keybias,keyxtalk,keyslot,pstatus = \
	    saltkey.instrumid(struct,infile,logfile)

	 # how many amplifiers?

	 if (pstatus == 0): 
	    nccds, pstatus = saltkey.get('NCCDS',struct[0],infile,logfile)
	    amplifiers = nccds * 2
	    if (nccds != 2):
		    message='SLOTMERGE-- ERROR:  Can not currently handle more than two CCDs'
		    pstatus=saltio.err(logfile, message)

	 # CCD geometry coefficients

	 if (pstatus == 0 and (instrume == 'RSS' or instrume == 'PFIS')):
	    xsh = [xshift[0], 0., xshift[1]]
	    ysh = [yshift[0], 0., yshift[1]]
	    rot = [rotation[0], 0., rotation[1]]
	    refid = 1
	 if (pstatus == 0 and instrume == 'SALTICAM'):
	    xsh = [xshift[0], 0.]
	    ysh = [yshift[0], 0.]
	    rot = [rotation[0], 0]
	    refid = 1

	 # how many extensions?

	 if (pstatus == 0): nextend, pstatus = saltkey.get('NEXTEND',struct[0],infile,logfile)

	 # how many exposures

	 if (pstatus == 0): exposures = nextend/amplifiers

	 # CCD on-chip binning

	 if (pstatus == 0): 
	    xbin, ybin, pstatus = saltkey.ccdbin(struct[0],infile,logfile)
	    gp = int(gap / xbin)

	 # create output hdu structure

	 if (pstatus == 0): 
	    outstruct = [None] * (exposures+1)
	    outstruct[0]=struct[0]

	 # iterate over exposures, stitch them to produce file of CCD images

	 for i in range(exposures):
	  if (pstatus == 0):

	      #Determine the total size of the image 
	      xsize=0
	      ysize=0
	      for j in range(amplifiers):
		hdu=i*amplifiers+j+1
		try:
		  xsize += len(struct[hdu].data[0])
		  if ysize < len(struct[hdu].data):
		     ysize=len(struct[hdu].data)
		except:
		  message='SLOTMERGE -- ERROR:  Unable to access extension %i ' % hdu
		  pstatus=saltprint.err(logfile,message)
		
	      if pstatus ==0:
	        xsize += gp* (nccds-1)
	        maxxsh, minxsh = determineshifts(xsh)
	        maxysh, minysh = determineshifts(ysh)
	        xsize += (maxxsh-minxsh)
	        ysize += (maxysh-minysh)

	      #Determine the x and y origins for each frame
	      if (pstatus ==0):
		 xdist=0
		 ydist=0
		 shid=0
		 x0=numpy.zeros(amplifiers)
		 y0=numpy.zeros(amplifiers)
	         for j in range(amplifiers):
		   x0[j]=xdist+xsh[shid]-minxsh
		   y0[j]=ysh[shid]-minysh
		   hdu=i*amplifiers+j+1
		   darr=struct[hdu].data
		   xdist += len(darr[0])
		   if j%2==1:
		      xdist += gp
		      shid += 1
		
	     
	      #make the out image     
	      if (pstatus == 0):  outarr = numpy.zeros((ysize, xsize), numpy.float64)
	     

	      #Embed each frame into the output array
	      if (pstatus == 0):
	         for j in range(amplifiers):
		   hdu=i*amplifiers+j+1
		   darr=struct[hdu].data
		   outarr, estatus=salttran.embed(darr, x0[j], y0[j], outarr, logfile)

	      #Add the outimage to the output structure
	      if (pstatus==0):
		 hdu=i*amplifiers+1
		 outhdu=i+1
		 outstruct[outhdu] = pyfits.ImageHDU(outarr)
		 outstruct[outhdu].header=struct[hdu].header
		

	      #Fix the headers in each extension
	      if (pstatus==0):
		 datasec='[1:%4i,1:%4i]' % (xsize, ysize)
                 pstatus = saltkey.put('DATASEC',datasec, outstruct[outhdu], outfile, logfile)
		 pstatus=  saltkey.rem('DETSIZE',outstruct[outhdu],outfile,logfile)
		 pstatus=  saltkey.rem('DETSEC',outstruct[outhdu],outfile,logfile)
		 pstatus=  saltkey.rem('CCDSEC',outstruct[outhdu],outfile,logfile)
		 pstatus=  saltkey.rem('AMPSEC',outstruct[outhdu],outfile,logfile)

	      #add housekeeping key words
	      if (pstatus==0):
	         outstruct[outhdu],pstatus=addhousekeeping(outstruct[outhdu], outhdu, outfile, logfile)

	      #add WCS keywords
	      #if (pstatus==0):
	         #outstruct[outhdu], pstatus=addWCS(outstruct[outhdu], xbin, ybin, outfile, logfile)	 
	

	      #print the information about what image has been merged
	      if (pstatus == 0):
		    message =  'SLOTMERGE  -- ' 
		    #saltprint.log(logfile,message,verbose)



         # close input FITS file

	 if (pstatus == 0): pstatus = saltio.closefits(struct,logfile)

        # housekeeping keywords

	 keymosaic='SLOTMERG'

	 if (pstatus == 0): pstatus = saltkey.put('SAL-TLM',time.asctime(time.localtime()),
					       outstruct[0],outfile,logfile)
	 if (pstatus == 0): pstatus = saltkey.new(keymosaic,time.asctime(time.localtime()),
					       'Amplifiers have been mosaiced',outstruct[0],outfile,logfile)
	 if (pstatus == 0): 
                history = 'SLOTMERGE -- '
                history += 'images=%s ' % infile
                history += 'outimages=%s ' % outfile
                history += history1
                pstatus = saltkey.history(outstruct[0],history,outfile,logfile)

	 #this is added for later use by 
         if (pstatus == 0): pstatus = saltkey.put('NCCDS', 0.5, outstruct[0], outfile, logfile)
         if (pstatus == 0): pstatus = saltkey.put('NSCIEXT', exposures, outstruct[0], outfile, logfile)
         if (pstatus == 0): pstatus = saltkey.put('NEXTEND', exposures, outstruct[0], outfile, logfile)

# clobber old files
	    
	 if (pstatus == 0 and clobber and os.path.isfile(outfile)): 
	    pstatus = saltio.delete(outfile,False,logfile)

# write FITS file of mosaiced image

	 if (pstatus == 0): outstruct = pyfits.HDUList(outstruct)
	 if (pstatus == 0): outstruct.writeto(outfile)
	 if (pstatus == 0): 
	    message = 'SLOTMERGE  -- created ' + os.path.basename(outfile) + ' at'
	    saltprint.time(message,logfile,verbose)
	 ifil += 1

# end time

    if (status == 0):
	saltprint.time('SLOTMERGE  -- completed at',logfile,verbose)
    else:
	saltprint.time('\nSLOTMERGE  -- aborted at',logfile,verbose)

#-----------------------------------------------------------
# Determine the maximum shifts in the data

def determineshifts(shift):
    maxsh=0
    minsh=0
    for sh in shift:
	 if sh > maxsh: maxsh=sh
	 if sh < minsh: minsh=sh
    return maxsh, minsh

#-----------------------------------------------------------
# WCS keywords

def addWCS(struct, xbin, ybin, outfile, logfile):

    status=0
    if (status == 0): status = saltkey.new('CRPIX1',1,'WCS: X reference pixel',
				       struct,outfile,logfile)
    if (status == 0): status = saltkey.new('CRPIX2',1,'WCS: Y reference pixel',
				       struct,outfile,logfile)
    if (status == 0): status = saltkey.new('CRVAL1',float(xbin), 'WCS: X reference coordinate value',
				       struct,outfile,logfile)
    if (status == 0): status = saltkey.new('CRVAL2',float(ybin), 'WCS: Y reference coordinate value',
				       struct,outfile,logfile)
    if (status == 0): status = saltkey.new('CDELT1',float(xbin),'WCS: X pixel size',
				       struct,outfile,logfile)
    if (status == 0): status = saltkey.new('CDELT2',float(ybin),'WCS: Y pixel size',
				       struct,outfile,logfile)
    if (status == 0): status = saltkey.new('CTYPE1','pixel','X type',
				       struct,outfile,logfile)
    if (status == 0): status = saltkey.new('CTYPE2','pixel','Y type',
				       struct,outfile,logfile)
    return struct, status

#-----------------------------------------------------------
# housekeeping keywords

def addhousekeeping(struct, outhdu, outfile, logfile):
    status=0
    if (status == 0): status = saltkey.put('SAL-TLM',time.asctime(time.localtime()), \
				       struct,outfile,logfile)
    if (status == 0): status = saltkey.new('EXTNAME','SCI','Extension name',         \
				       struct,outfile,logfile)
    if (status == 0): status = saltkey.new('EXTVER',outhdu,'Extension number',       \
				       struct,outfile,logfile)

    return struct, status

# -----------------------------------------------------------
# main code 

parfile = iraf.osfn("slottools$slotmerge.par") 
t = iraf.IrafTaskFactory(taskname="slotmerge",value=parfile,function=slotmerge, pkgname='slottools')
