#!/usr/bin/env python

# Author                 Version      Date
# -----------------------------------------------
# Martin Still (SAAO)    0.2          13 Nov 2006
# S M Crawford (SAA0)    0.3          11 Feb 2008
# S M Crawford (SAA0)    0.31         16 Jul 2008
#
# SLOTPHOT is a tool to do photometry on a serious of data 
# obtained with slotmode configuration of salticam.  
# After reading in the arguments, the algorithm is the
# following:
# 1.  Open th fits file
# 2.  Subtact a smooth background  
# 3.  Centroid the comparison star to acount for drift
# 4.  Photometry on Comparison and target star
# 5.  Write data
# 6.  Plot data [optional]
# 7.  Write fits [optional]
#
# One of the basic assumptions of this program are that 
# the target and comparison star are on the same amplifier
#
# Updates:
# ----------------------------------------------------------
# 20080616--Fixed a number of small bugs
#         --Updated the handling of multiple amplifiers so that
#           the program can now handle having the newfits as
#	    as an input file
#
# 20080910--Corrected the background subtraction.  Now a pixel
#           is replaced with the median of its row if its two bright
#         --Added option to allow row-by-row median subtraction
#
# 20081027--Added limits to the finddrift to limit it to within a 
#           set number of pixels
#         --Gives the users on local background subraction and annullur
#           definition


import os,sys, time, re
from pyraf import iraf 
import saltprint, saltkey, saltio, salttime
import slottool
from slotbackground import subbackground

import numpy
import pyfits

# ----------------------------------------------------------
# core routine

def slotphot(images,outfile,srcfile, newfits, phottype, subbacktype, sigback, \
             mbin, sorder,  iter, sigdet, contpix, ampperccd, ignorexp, driftlimit, \
             finddrift, reltime, clobber,logfile,verbose,status):



#set up the variables
    status = 0
    entries = []
    vig_lo = {}
    vig_hi = {}
    amp = {}
    x = {}
    y = {}
    x_o = {}
    y_o = {}
    r = {}
    br1 = {}
    br2 = {}
    hour = 0
    min = 0
    sec = 0.
    time0 = 0.
    nframes = 0
    bin=mbin
    order=sorder
    plotdata=False
    plotfreq=0.01
    


# test the logfile

    logfile = saltio.logname(logfile)

# log the call 

    saltprint.line(logfile,verbose)
    message = 'SLOTPHOT -- '
    message += 'images='+images+' '
    message += 'outfile='+outfile+' '
    message += 'srcfile='+srcfile+' '
    message += 'newfits='+newfits+' '
    message += 'subbacktype='+subbacktype+' '
    message += 'phottype='+phottype+' '
    message += 'sigback=%f ' % sigback
    message += 'bin=%i ' % bin
    message += 'iter=%i '% iter
    message += 'order=%i ' % order
    message += 'sigdet=%f ' % sigdet
    message += 'contpix=%i ' % contpix
    message += 'ampperccd=%i ' % ampperccd
    message += 'ignorexp=%i ' % ignorexp 
    message += 'driftlimit=%f ' % driftlimit
    yn = 'n'
    if (finddrift): yn = 'y'
    message += 'finddrift='+yn+' '
    yn = 'n'
    if (reltime): yn = 'y'
    message += 'reltime='+yn+' '
    yn = 'n'
    if (clobber): yn = 'y'
    message += 'clobber='+yn+' '
    message += 'logfile='+logfile+' '
    yn = 'n'
    if (verbose): yn = 'y'
    message += 'verbose='+yn+'\n'
    saltprint.log(logfile,message,verbose)

# start time

    saltprint.time('SLOTPHOT started at  ',logfile,verbose)

# is the input file specified?
   
    if (status == 0): status = saltio.filedefined('Input',images,logfile)

# if the input file is a list, does it exist?

    if (status == 0 and images[0] == '@'): 
        status = saltio.listexists('Input',images,logfile)

# parse list of input files

    if (status == 0):
        infiles, status = saltio.listparse('Raw image',images,'','','',logfile)

# check input files exist

    if (status == 0): status = saltio.filesexist(infiles,'','r',logfile)

# is the output file specified?
   
    if (status == 0): status = saltio.filedefined('Output',outfile,logfile)

# check output file does not exist, optionally remove it if it does exist

    if (status == 0): 
	status = saltio.filedoesnotexist(outfile,'no',logfile)
        if (status == 1 and clobber): 
	  os.remove(outfile)
	  status = 0
        elif (status == 1 and not clobber):
	  message = 'ERROR -- SLOTPHOT: File ' + outfile + ' already exists, use clobber=y'
	  status = saltprint.err(logfile,message)

# open output ascii file

    if (status == 0):
	try:
	    lc = open(outfile,'a')
	except:
	    message = 'ERROR -- SLOTPHOT: Cannot open ouput file ' + outfile
	    status = saltprint.err(message,logfile)

# is the extraction region defintion file specified?
   
    if (status == 0): status = saltio.filedefined('Extraction region defintion',srcfile,logfile)

# check extraction region defintion file exists

    srcfile = srcfile.strip()
    if (status == 0): status = saltio.fileexists(srcfile,logfile)

# read extraction region defintion file

    if (status == 0):
	region, status = saltio.openascii(srcfile,'r',logfile)
	for line in region:
	    if (len(line.strip()) > 0 and line[0] != '#'):
		line = line.rstrip('\r\n')
		line = line.lstrip()
		line = re.sub("\s+",",",line)
		line.rstrip(',')
		entries = line.split(',')
		if (entries[0] == '1' or entries[0] == 't'): entries[0] = 'target'
		if (entries[0] == '2' or entries[0] == 'c'): entries[0] = 'comparison'
		if (entries[0] == 'target' or entries[0] == 'comparison'):
		    amp[entries[0]] = int(entries[1])
		    x[entries[0]] = int(float(entries[2]))
		    y[entries[0]] = int(float(entries[3]))
		    x_o[entries[0]] = int(float(entries[2]))
		    y_o[entries[0]] = int(float(entries[3]))
		    r[entries[0]] = int(float(entries[4]))
		    br1[entries[0]] = int(float(entries[5]))
		    br2[entries[0]] = int(float(entries[6]))
	status = saltio.closeascii(region,logfile)

# is extraction region data complete

    if status==0:
     if (len(amp) != 2 or len(x) != 2 or len(y) != 2 or len(r) != 2 or len(br1) != 2 or len(br2) != 2):
	message = 'ERROR: SLOTPHOT -- extraction region definition file '+srcfile
	message += ' does not have an acceptable format.'
	status = saltprint.err(logfile,message)

# set the writenewfits parameter

    if status==0:
	if not newfits or newfits=='none':
	    writenewfits=False
	else:
	    writenewfits=newfits


# get time of first exposure and basic information about the observations


    if (status == 0):

	infile=infiles[0]
	struct, status = saltio.openfits(infile,logfile)

        # identify instrument
        if (status ==0 ):
	   instrume,keyprep,keygain,keybias,keyxtalk,keyslot,status = saltkey.instrumid(struct,infile,logfile)

	# how many extensions?
    	if (status == 0): 
		nextend,status = saltkey.get('NEXTEND',struct[0],infile,logfile)
		if nextend < amp['comparison']:
		   message='SALTPHOT--ERROR:  Insufficient number of extensions in %s' % (infile)
		   status=saltprint.err(logfile, message)

	# how many amplifiers?
    	if (status == 0): 
	   amplifiers,status = saltkey.get('NCCDS',struct[0],infile,logfile)
	   amplifiers = int(ampperccd*float(amplifiers))
	   if ampperccd>0:
	      nframes = nextend/amplifiers
	      nstep=amplifiers
	   else:
	      nframes = nextend
	      nstep=1
	   ntotal=nframes*len(infiles)

	# image size
        if (status == 0):
	   naxis1,status = saltkey.get('NAXIS1',struct[amp['comparison']],infile,logfile) 
	   naxis2,status = saltkey.get('NAXIS2',struct[amp['comparison']],infile,logfile) 

	# CCD binning
        if (status == 0):
	  ccdsum,status = saltkey.get('CCDSUM',struct[0],infile,logfile) 
	  binx = int(ccdsum.split(' ')[0])
	  biny = int(ccdsum.split(' ')[1])


	#Identify the time of the observations
	hdu = 1
	try:
	  time0=slottool.getobstime(struct[hdu], infile+'['+str(hdu)+']', logfile)
	  dateobs, status = saltkey.get('DATE-OBS',struct[hdu],infile,logfile)
	  dateobs = dateobs.replace('-','/')
	except:
	   message='ERROR--SLOTPHOT:  No time or obsdate in first image'
	   status=saltprint.err(logfile,message)

        # If a total file is to written out, create it and update it
        if status==0 and writenewfits:
	   #delete the file if clobber is on and the file exists
	   if (status == 0 and os.path.isfile(writenewfits)):
	     if clobber:
              status = saltio.delete(writenewfits,'no',logfile)
	     else:
	      message='ERROR -- SLOTPHOT: Output fits file ' + writenewfits + 'already exists. '
	      message += 'Use Clobber=yes to overwrite file'
              status = saltprint.err(logfile,message)
	
	   #create the output file 
	   if status==0:
	    try:
              hdu = pyfits.PrimaryHDU()
              hdu.header=struct[0].header
	      hdu.header['NCCDS']=1
	      hdu.header['NSCIEXT']=ntotal-ignorexp
	      hdu.header['NEXTEND']=ntotal-ignorexp
              hduList = pyfits.HDUList(hdu)
              hduList.verify()
              hduList.writeto(writenewfits)
            except:
              message  = 'ERROR -- SLOTPHOT: Could not create new fits file named ' + writenewfits
              status = saltprint.err(logfile,message)
	
	   #read it back in for updating
	   if status==0:
            try:
                  hduList = pyfits.open(writenewfits,mode='update')
            except:
                  message = 'ERROR -- SLOTPHOT:  Canot open  ' + outfile
                  status = saltprint.err(logfile,message)


    # set up the arrays
    if status==0:
        j=0
        x2=float(j)/float(ntotal)
        ftime=plotfreq
	time=numpy.zeros(ntotal ,dtype='float')-1.0
	dx=numpy.zeros(ntotal ,dtype='float')
	dy=numpy.zeros(ntotal ,dtype='float')
	tflux=numpy.zeros(ntotal ,dtype='float')
	terr =numpy.zeros(ntotal ,dtype='float')
	cflux=numpy.zeros(ntotal ,dtype='float')
	cerr =numpy.zeros(ntotal ,dtype='float')
	ratio=numpy.zeros(ntotal ,dtype='float')
	rerr =numpy.zeros(ntotal ,dtype='float')

    #set up the completeness tracker
    if status==0 and verbose:
       ctext='Percentage Complete: %3.2f' % x2
       sys.stdout.write(ctext)
       sys.stdout.flush()


    if status==0:
      for infile in infiles:

	if (status == 0 ): struct, status = saltio.openfits(infile,logfile)

        # has file been cleaned?
        if (status == 0): cstatus=saltkey.clean(struct,infile,keyslot,keygain,logfile)

	#Skip through the frames and process each frame individually
	for i in range(nframes):
          if not (infile==infiles[0] and i < ignorexp):
	   pstatus=0
	   hdu=amp['comparison']+i*nstep 
	   try:
	     header=struct[hdu].header
	     array=struct[hdu].data
	     array=array*1.0
	   except:
	     message='SLOTPHOT--ERROR:  Unable to open extension %i in image %s' % (hdu, infile)
	     pstatus=saltprint.err(logfile,message)


	   #start the analysis of each frame
	   if pstatus==0:
             # get the time
	     time[j]=slottool.getobstime(struct[hdu],infile+'['+str(hdu)+']',logfile)

             # gain and readout noise
	     try:
	       gain=float(header['GAIN'])
	     except:
	       gain=1
	       message='WARNING--SLOTPHOT:  Gain not specified in image header'
	       saltprint.log(logfile,message,verbose)
	     try:
	       rdnoise=float(header['RDNOISE'])
	     except:
	       rdnoise=0
	       message='WARNING--SLOTPHOT:  RDNOISE not specified in image header'
	       saltprint.log(logfile,message,verbose)

	   # background subtraction
	   if pstatus==0 and not subbacktype=='none':
	     array,status = subbackground(array, sigback, bin, order, iter, subbacktype, logfile)

	   #x-y fit to the comparison star and update the x,y values
	   if pstatus==0 and finddrift:
	     carray, fx,fy,dstatus = slottool.finddrift(array, x['comparison'], y['comparison'], r['comparison'], \
					naxis1, naxis2, sigdet, contpix, sigback, driftlimit, iter, logfile)
	     if fx > -1  and fy > -1:
              if fx < naxis1 and fy < naxis2:
	       dx[j]=x['comparison']-fx
	       dy[j]=y['comparison']-fy
	       x['comparison']=fx
	       y['comparison']=fy
	       x['target']=x['target']-dx[j]
	       y['target']=y['target']-dy[j]
	      else:
	       dx[j]=0
	       dy[j]=0
	       x['comparison']=x_o['comparison']
               y['comparison']=y_o['comparison']
               x['target']=x_o['target']
               y['target']=y_o['target']
	     else:
	      message  = 'WARNING--SLOTPHOT:  No comparison object found in image file ' + infile
	      message += ' on extension %i \n' % hdu
	      dstatus=saltprint.err(logfile, message)

	
	   #do photometry
	   if pstatus==0:
		
	     tflux[j],terr[j],cflux[j],cerr[j],ratio[j],rerr[j],status =  \
		slottool.dophot(phottype, array, x, y, r, br1, br2, gain, rdnoise, naxis1, naxis2, verbose, logfile)

	   #record results
	   if pstatus==0:
             status=slottool.writedataout(lc, j+1, time[j], x, y, tflux[j], terr[j], cflux[j],cerr[j],ratio[j],rerr[j], \
		time0, reltime, logfile, verbose)
	
	   #plot data 
	   if pstatus==0 and verbose:
	     x2=float(j)/float(ntotal)
	     if x2 > ftime:
              ctext='\b\b\b\b\b %3.2f' % x2
              sys.stdout.write(ctext)
              sys.stdout.flush()
	      ftime += plotfreq
	
	   #write the fits file
	   if pstatus==0 and writenewfits: 
		hdue = pyfits.ImageHDU(array)
		hdue.header=header
		#update the header values with the image name and extension number
		try:
		     hdue.header.update('ONAME',infile,'Original image name')
		     hdue.header.update('OEXT',hdu,'Original extension number')

		except:
		     message='SALTPHOT--WARNING:  Could not update image in newfits file for %s ext %i' % (infile, hdu)
	             saltprint.log(logfile,message,verbose)
		hduList.append(hdue)

	   j +=1
	   if status==1: print j

        # close FITS file		
	status = saltio.closefits(struct,logfile)

    print('')

# close the output fits file:
	
    if status==0 and writenewfits:
	  try:
	   #update the header with the correct number of extensions

	   #write out the file
           hduList.flush()
           hduList.close()
          except:
           message = 'ERROR: SLOTPHOT-- Failed to write ' + writenewfits
           status = saltprint.err(logfile,message)


# close output ascii file

    if (status == 0):
	try:
	    lc.close()
	except:
	    message = 'ERROR -- SLOTPHOT: Cannot close ouput file ' + outfile
	    status = saltprint.err(message,logfile)

# end time

    if (status == 0):
	saltprint.time('SLOTPHOT completed at',logfile,verbose)
    else:
	saltprint.time('SLOTPHOT aborted at',logfile,verbose)


# -----------------------------------------------------------
# Plot the data 

def makeplotdata(x2, logfile, verbose): 
    """As the data is measured, plots the target and companion, the drift, both light curves and the ratio
   
       returns status
    """

    status=0
    #textplot
    text_plot=axes([0.1,0.55,0.8,0.05], frameon=False)
    text_plot.cla()
    ctext='Percentage Complete: %3.2f' % x2
    text(0.33, 0.2, ctext)
    setp(gca(),xticks=[],yticks=[])
    text_plot.set_xlim(0,1)
    text_plot.set_ylim(0,1)
    #plot the completeness graphux curve
    comp_plot = axes([0.1,0.5,0.8,0.05])
    comp_plot.cla()
    hold(True)
    x=[0,x2,x2,0]
    y=[0,0,1,1]
    fill(x,y,ec='b',fc='b')
    setp(gca(),xticks=[],yticks=[])
    comp_plot.set_xlim(0,1)
		

    #plot both flux curves
    ion()
    ylabel('')
    ioff()

    return status


# -----------------------------------------------------------
# main code 

parfile = iraf.osfn("slottools$slotphot.par") 
t = iraf.IrafTaskFactory(taskname="slotphot",value=parfile,function=slotphot,pkgname='slottools')
