# Author                 Version      Date
# -----------------------------------------------
# S M Crawford (SAA0)    0.1          20 Feb 2008
#
# slotphottools contains the algorithms to perform 
# photometry on slotmode data.  

from pyraf import iraf 
import saltprint, saltkey, saltio, salttime, saltstat

from numpy import *

# ----------------------------------------------------------
# Calculate the image statistics 

def calcstat(image, sig, iter):
    imsize=size(image)
    if imsize<=0: return 0

    mean=average(image)
    stddev=sqrt(sum((image-mean)**2)/imsize)
    for i in range(iter):
	mean=saltstat.mean2dclip(image,mean, stddev, sig)
	stddev=saltstat.std2dclip(image,mean, stddev, sig)
    return mean,stddev


# ----------------------------------------------------------
# Check to see that the boundary conditions are okay

def checkedge(x, x0, x1):
    """Determeine if the value x is over the edge
	of an image
    """
    x=int(x)
    if x < x0: return x0
    if x > x1: return x1
    return x

# ----------------------------------------------------------
# Find the x-y centroid in a region 

def findcentroid(array):
    """ Find the centroid of an array where
	cx=sqrt(sum(f*x*x)/sum(f))
	
	returns cx,cy
    """

    cx=-1
    cy=-1

    try:
      i=len(array)
      j=len(array[0])
    except:
      return cx, cy 

    y,x=indices((i,j))
    tot=array.sum()
    if tot!=0:
      cx=(array*x*x).sum()/tot
      cy=(array*y*y).sum()/tot

    if cx > 0: cx=sqrt(cx)
    if cy > 0: cy=sqrt(cy)

    return cx,cy

# -----------------------------------------------------------
# Find significant pixels

def objdetect(image, sigdet, stddev):
    """Detect all significant pixels in an image
       returns count
    """
    mask = (image > sigdet*stddev)
    return mask.sum()

# -----------------------------------------------------------
# Given a central x-y, calculate the drift 

def calcdrift(image, x, y, r, naxis1, naxis2):
    x1=checkedge(x-r,0,naxis1)
    x2=checkedge(x+r,0,naxis1)
    y1=checkedge(y-r,0,naxis2)
    y2=checkedge(y+r,0,naxis2)
    cimage=image[y1:y2,x1:x2]
    fx,fy=findcentroid(cimage)
    if fx >= 0 and fy >= 0:
     cx=x1+fx
     cy=y1+fy
    else:
     cx=-1
     cy=-1
    return cimage, cx, cy

# -----------------------------------------------------------
# Find the x-y drift of the comparison star

def findneareststar(array, xc, yc, rc, std, dsig, cpix, naxis1, naxis2):
    """Search an image for all stars and find the nearest one to the
       given coordinates

       returns an x,y
    """
    import scipy.ndimage as nd
    fx=-1
    fy=-1
 
    #detect objects and find their positions
    mask = (array > dsig*std)
    obj_labels, obj_num = nd.label(mask)
    if obj_num<1: return fx,fy

    coolist=nd.center_of_mass(array,obj_labels,range(obj_num))
 
    #find the closest coordinate to the previous coordinate
    if obj_num>1:
     best_rad=10000
     for y,x in coolist:
       rad=sqrt((x-xc)**2+(y-yc)**2)
       if rad < best_rad:
	  best_rad=rad
	  fx=x
	  fy=y
    else:
	fy=coolist[0]
	fx=coolist[1]

    #Confirm the source is a good source
    if fx > -1 and fy > -1:
      carray, fx, fy = calcdrift(array, fx, fy, rc, naxis1, naxis2)
      cimsize=size(carray)
      if cimsize < cpix:  
	fx=-1
	fy=-1

    return fx,fy 


# -----------------------------------------------------------
# Find the x-y drift of the comparison star

def finddrift(image, cx, cy, cr, naxis1, naxis2, sigdet, cpix,  sigback, driftlimit, iter, log):
    """Starting with the initial x-y of the comparison source, search for 
       the comparison star and see how much it has drifted.  Expanding the 
       search if no comparison star is found
  
       returns dx,dy,status
    """
    fx=-1
    fy=-1
    status=0

    #set the image statistics for the image
    mean,stddev=calcstat(image,sigback,iter)

    #First to the fast search for the object, assuming it is roughly in the same position
    if driftlimit>0: 
	dr=int(driftlimit)
    else:
	dr=cr
    cimage, cx2, cy2 = calcdrift(image, cx, cy, dr, naxis1, naxis2)
    cimsize=size(cimage)
    nobj=objdetect(cimage, sigdet, stddev)
    if cx2 >= 0 and cy2 >= 0 and nobj > cpix:
        fx=cx2
        fy=cy2
        #repeat several times to grab the best x-y
        for i in range(iter):
         cimage, fx, fy = calcdrift(image, fx, fy, dr, naxis1, naxis2)
        return cimage, fx, fy, status

    

    #set up the image to be searched
    if driftlimit>0:
       x1=checkedge(cx-dr,0,naxis1)
       x2=checkedge(cx+dr,0,naxis1)
       y1=checkedge(cy-dr,0,naxis2)
       y2=checkedge(cy+dr,0,naxis2)
       cimage=image[y1:y2,x1:x2]
    else:
       cimage=image.copy()

    cx2,cy2=findneareststar(cimage,cx,cy,dr,stddev, sigdet, cpix, naxis1,naxis2)
    if cx2 >= 0 and cy2 >= 0:
        fx=cx2+x1
        fy=cy2+y1
        #repeat several times to grab the best x-y
        for i in range(iter):
         cimage, fx, fy = calcdrift(image, fx, fy, cr, naxis1, naxis2)
        return cimage, fx, fy, status 
     
    #if nothing was found, return values of -1
    return cimage, fx,fy,status

# -----------------------------------------------------------
# get time since noon the previous day

def getobstime(struct, infile, logfile):
      """Return the time since noon the previous day in seconds
      """
      time=-1
      hour,min,sec,status = saltkey.timeobs(struct,infile,logfile)
      time = salttime.utc2sec(hour,min,sec)
      if (hour > 12):
		time -= 43200.
      else:
		time += 43200.

      return time

# -----------------------------------------------------------
# Determine the background value

def calcstarbackground(array, x, y, br1, br2, gain, rdnoise, naxis1, naxis2, logfile):
    """Determines the sky background around a single star.  Using
     square annulli.  returns the background flux in terms of e/pix
  
     returns bflux, npix, status
    """
    status=0
    #calculate counts from first annullus
    bx1=checkedge(x-br1,0,naxis1)
    bx2=checkedge(x+br1,0,naxis1)
    by1=checkedge(y-br1,0,naxis2)
    by2=checkedge(y+br1,0,naxis2)
    bflux1=array[by1:by2,bx1:bx2].sum()
    barea1=size(array[by1:by2,bx1:bx2])
 
    #calculate counts from second annullus
    bx1=checkedge(x-br2,0,naxis1)
    bx2=checkedge(x+br2,0,naxis1)
    by1=checkedge(y-br2,0,naxis2)
    by2=checkedge(y+br2,0,naxis2)
    bflux2=array[by1:by2,bx1:bx2].sum()
    barea2=size(array[by1:by2,bx1:bx2])
    npix=barea2-barea1

    #if enough pixels, calculate the sky background
    if npix>0:
	bflux=(bflux2-bflux1)/(barea2-barea1)
	berr=sqrt(abs(bflux)+npix*rdnoise**2)/npix
    else:
        bflux=0
	berr=0
	message= "ERROR--SALTPHOT: Poor Background area specified.  Make sure br2 > br1"
	status=saltprint.err(logfile, message)

    return bflux, berr, npix, status


# -----------------------------------------------------------
# perform square aperture phometry 

def sqapphot(array, xc, yc, rc, br1, br2,gain, rdnoise,  naxis1, naxis2, logfile):
    """Perform square aperture photometry

       returns flux, flux_err, status
    """
    status =0

    #first calculate the flux
    x1=checkedge(xc-rc,0,naxis1)
    x2=checkedge(xc+rc,0,naxis1)
    y1=checkedge(yc-rc,0,naxis2)
    y2=checkedge(yc+rc,0,naxis2)
    area=size(array[y1:y2,x1:x2])
    if area==0:
        flux=0
	flux_err=0
	message= "ERROR--SALTPHOT: No data in the specified array"
	status=saltprint.err(logfile, message)

    if status==0:
     flux=array[y1:y2,x1:x2].sum()
     #calculate the background flux
     bflux, berr, npix, status = calcstarbackground(array, xc, yc, br1, br2, gain, rdnoise, naxis1, naxis2, logfile)

    #calculate the flux
    if status==0:
     flux=flux-area*bflux
     flux_err=sqrt(abs(flux)+area*rdnoise**2+area*berr**2+area**2*berr**2/npix)
    return  flux, flux_err, status

# -----------------------------------------------------------
# perform circular aperture phometry

def capphot(array, xc, yc, rc, br1, br2,gain, rdnoise,  naxis1, naxis2, logfile):
    """Perform circular aperture photometry
	 
       return flux, flux_err, status
    """

    status =0

    #calculate the flux
    try:
     i=len(array)
     j=len(array[0])
     y,x=indices((i,j))
     r=sqrt((x-xc)**2+(y-yc)**2)
     mask=(r<rc)
     area=mask.sum()
    except:
	area=0

    if area==0:
        flux=0
	flux_err=0
	message= "ERROR--SALTPHOT: No data in the specified array"
	status=saltprint.err(logfile, message)

    if status==0:
     flux=(array*mask).sum()
     #calculate the background flux
     bflux, berr, npix, status = calcstarbackground(array, xc, yc, br1, br2, gain, rdnoise, naxis1, naxis2, logfile)

    #calculate the flux
    if status==0:
     flux=flux-area*bflux
     flux_err=sqrt(abs(flux)+area*rdnoise**2+area*berr**2+area**2*berr**2/npix)
    return  flux, flux_err, status

# -----------------------------------------------------------
# perform curve of growth photometry 

def cogphot(array, xc, yc, rc, br1, br2,gain, rdnoise,  naxis1, naxis2, logfile):
    """Perform curve of growth photometry.  Will use <rc> square apertures between 2 pix
       to the br2 radius.  rc now gives the number of apertures to use.  br1 gives 
       the annulli to use for sky determination.  The flux returned is the flux in the
       aperture with the highest signal to noise

       return flux, flux_err, status
    """

    status=0
    min_rad=2
    rad=zeros(rc,dtype='float')
    flux=zeros(rc,dtype='float')
    flux_err=zeros(rc,dtype='float')
    rstep=(br2-min_rad)/rc

    best_sn=0
    best_i=-1
    for i in range(rc):
     if status==0:
	rad[i]=min_rad+i*rstep
        flux[i], flux_err[i], status=sqapphot(array, xc, yc, rad[i], br1, br2, gain, rdnoise, naxis1, naxis2, logfile)
	s_n=flux[i]/flux_err[i]
	if s_n > best_sn:
	   best_sn=s_n
	   best_i=i

    # if no value is found, use the middle aperture
    if best_i==-1: best_i=int(len(rad)/2)

    return rad[best_i], flux[best_i], flux_err[best_i], status

# -----------------------------------------------------------
# calculate the optimal ratio  

def calc_optimal(array, xt, yt, xc, yc, rc, br1, br2,gain, rdnoise,  naxis1, naxis2, logfile):
    """Calculate the optimal ratio between the two sources based X^2 pixel fit of the 	
	comparison star to the object star.  For both stars the background will be removed
	and then the comparison star will be used as a 'model' psf for the X^2 fit.  
	To simplify the error calculation, we are assume the errors on the comparison
	star are much smaller than the errors on the target star

       return ratio, ratio_err, status
    """
    status=0
    ratio=0
    ratio_err=0
    #find the background for both images
    btflux, bterr, ntpix, status = calcstarbackground(array, xt, yt, br1, br2, gain, rdnoise, naxis1, naxis2, logfile)
    bcflux, bcerr, ncpix, status = calcstarbackground(array, xc, yc, br1, br2, gain, rdnoise, naxis1, naxis2, logfile)

    #create the target array
    x1=checkedge(xt-rc,0,naxis1)
    x2=checkedge(xt+rc,0,naxis1)
    y1=checkedge(yt-rc,0,naxis2)
    y2=checkedge(yt+rc,0,naxis2)
    tarray=array[y1:y2,x1:x2]
    tarea=size(tarray)
    tarray=tarray-btflux

    #create the comparison array
    x1=checkedge(xc-rc,0,naxis1)
    x2=checkedge(xc+rc,0,naxis1)
    y1=checkedge(yc-rc,0,naxis2)
    y2=checkedge(yc+rc,0,naxis2)
    carray=(array[y1:y2,x1:x2])
    carea=size(carray)
    carray=carray-bcflux

    #check to make sure the objects are the same size
    if size(tarray)>size(carray):
        x1=checkedge(xc-rc,0,naxis1)
        x2=checkedge(xc+rc,0,naxis1)
        y1=checkedge(yc-rc,0,naxis2)
        y2=checkedge(yc+rc,0,naxis2)
	dx1=int(xc-x1)
	dx2=int(x2-xc)
	dy1=int(yc-y1)
	dy2=int(y2-yc)
        x1=checkedge(xt-dx1,0,naxis1)
        x2=checkedge(xt+dx2,0,naxis1)
        y1=checkedge(yt-dy1,0,naxis2)
        y2=checkedge(yt+dy2,0,naxis2)
        tarray=array[y1:y2,x1:x2]
        tarea=size(tarray)
        tarray=tarray-btflux
    if size(tarray)<size(carray):
        x1=checkedge(xt-rc,0,naxis1)
        x2=checkedge(xt+rc,0,naxis1)
        y1=checkedge(yt-rc,0,naxis2)
        y2=checkedge(yt+rc,0,naxis2)
	dx1=int(xt-x1)
	dx2=int(x2-xt)
	dy1=int(yt-y1)
	dy2=int(y2-yt)
        x1=checkedge(xc-dx1,0,naxis1)
        x2=checkedge(xc+dx2,0,naxis1)
        y1=checkedge(yc-dy1,0,naxis2)
        y2=checkedge(yc+dy2,0,naxis2)
        carray=array[y1:y2,x1:x2]
        carea=size(carray)
        carray=carray-bcflux
	
    #check to make sure they are the same size
    #and fail if they are not
    if carea!=tarea:
	status=1
	return ratio,ratio_err,status

    #find the ratio between the two
    psum=(carray*carray).sum()
    if psum == 0:
      ratio = 0
      ratio_err=1
    else:
      ratio = (tarray*carray).sum()/psum
      tflux,terr, status=sqapphot(array, xt, yt, rc, br1, br2,gain, rdnoise,  naxis1, naxis2, logfile)
      cflux,cerr, status=sqapphot(array, xc, yc, rc, br1, br2,gain, rdnoise,  naxis1, naxis2, logfile)
      if not tflux==0 and not cflux==0:
        ratio_err=ratio*sqrt((terr/tflux)**2+(cerr/cflux)**2)
      else:
	ratio_err=0


    return ratio, ratio_err, status


# -----------------------------------------------------------
# Routine to perform the photometry

def dophot(phottype, array, x, y, r, br1, br2, gain, rdnoise, naxis1, naxis2, verbose, logfile):
    """Routine to perform photometry on the image.  It will perform the requested type of photometry
       (square aperture, circular aperture, curve of growth, psf fit, or optimal).  See the help file
       for descriptions of each
	
       returns list of photometry parameters for target and companion, status
    """
    status=0
    time=1

    if phottype=='square':
	target_flux, target_flux_err, status=sqapphot(array, x['target'], y['target'], r['target'], \
			br1['target'], br2['target'], gain, rdnoise,  naxis1, naxis2, logfile)
	compar_flux, compar_flux_err, status=sqapphot(array, x['comparison'], y['comparison'], r['comparison'], \
			br1['comparison'], br2['comparison'], gain, rdnoise,  naxis1, naxis2, logfile)
	ratio, ratio_err = calcratio(target_flux, target_flux_err, compar_flux, compar_flux_err)
    elif phottype=='circular' :
	target_flux, target_flux_err, status=capphot(array, x['target'], y['target'], r['target'], \
			br1['target'], br2['target'], gain, rdnoise,  naxis1, naxis2, logfile)
	compar_flux, compar_flux_err, status=capphot(array, x['comparison'], y['comparison'], r['comparison'], \
			br1['comparison'], br2['comparison'], gain, rdnoise,  naxis1, naxis2, logfile)
	ratio, ratio_err = calcratio(target_flux, target_flux_err, compar_flux, compar_flux_err)
    elif phottype=='cog':
	r_cog, compar_flux, compar_flux_err, status=cogphot(array, x['comparison'], y['comparison'], r['comparison'], \
			br1['comparison'], br2['comparison'], gain, rdnoise,  naxis1, naxis2, logfile)
	target_flux, target_flux_err, status=sqapphot(array, x['target'], y['target'], r_cog, \
			br1['target'], br2['target'], gain, rdnoise,  naxis1, naxis2, logfile)
	ratio, ratio_err = calcratio(target_flux, target_flux_err, compar_flux, compar_flux_err)
    elif phottype=='psf':
	compar_flux, compar_flux_err, status=sqapphot(array, x['comparison'], y['comparison'], r['comparison'], \
			br1['comparison'], br2['comparison'], gain, rdnoise,  naxis1, naxis2, logfile)
	target_flux, target_flux_err, status=sqapphot(array, x['target'], y['target'], r['target'], \
			br1['target'], br2['target'], gain, rdnoise,  naxis1, naxis2, logfile)
    elif phottype=='optimal':
	compar_flux, compar_flux_err, status=sqapphot(array, x['comparison'], y['comparison'], r['comparison'], \
			br1['comparison'], br2['comparison'], gain, rdnoise,  naxis1, naxis2, logfile)
	ratio, ratio_err, status = calc_optimal(array, x['target'], y['target'], x['comparison'], y['comparison'], r['comparison'], \
                        br1['comparison'], br2['comparison'], gain, rdnoise,  naxis1, naxis2, logfile)
	target_flux=ratio*compar_flux
	if not ratio==0 and not compar_flux:
	  target_flux_err=target_flux*sqrt((compar_flux_err/compar_flux)**2+(ratio_err/ratio)**2)
	else:
	  target_flux_err=compar_flux_err
    else:
	message='ERROR--SALTPHOT:  Photometry Type does not exist'
	status=saltprint.err(logfile,message)
	return 0,0,0,0,0,0, status
    #place the appropriate information into pinfo
    return target_flux, target_flux_err, compar_flux, compar_flux_err, ratio, ratio_err, status

# -----------------------------------------------------------
# Calculate the ratio for a simple case

def calcratio(target_flux, target_flux_err, compar_flux, compar_flux_err):
    """Calculate the flux ratio

	return ratio, ratio_err
    """
    if compar_flux>0:
	ratio=target_flux/compar_flux
    else:
	ratio=0

    if compar_flux > 0 and target_flux > 0:
     ratio_err=ratio*sqrt((target_flux_err/target_flux)**2+(compar_flux_err/compar_flux)**2)
    else:
     ratio_err=ratio*sqrt(target_flux_err**2+compar_flux_err**2)

    return ratio, ratio_err



# -----------------------------------------------------------
# Write the data out to a file

def writedataout(lout, index, time, x, y, target_flux, target_flux_err, compar_flux, compar_flux_err, \
		 ratio, ratio_err, time0, reltime, logfile, verbose):
    status=0
    #calculate the ratio of the target to comparison star
    target_x=x['target']
    target_y=y['target']
    compar_x=x['comparison']
    compar_y=y['comparison']

    #if reltime is selected, only write out the time since the first image
    if reltime:
	wtime=time-time0
    else:
	wtime=time

    #write out the statement
    statement = '%5i %10.3f %10.6f %7.6f %5.2f %5.2f %10.3f %7.3f %5.2f %5.2f %10.3f %7.3f  \n' % \
	 (index, wtime, ratio, ratio_err, target_x, target_y, target_flux, target_flux_err, compar_x, compar_y, compar_flux, compar_flux_err)
    try:
      lout.write(statement)
    except:
      message="ERROR-SALTPHOT:  Unable to write to the output file"
      status=saltprint.err(logfile,message)
    return status

