#!/usr/bin/env python

# Author                 Version      Date
# -----------------------------------------------
# S M Crawford (SAA0)    0.30         25 Oct 2008
#
# SLOTUTC is a tool to fix the UTC time in SALTICAM 
# slot mode imaging.  Because of a timing errror in 
# OS being used, the clock loses time between seconds 
# and the UTC time is inaccurate between updates of 
# the system clock (which is currently updated at the
# beginning of each second.
#
# As such this program reads in a set of slotmode data
# and calculates what the true exposure plus deadtime is.
# Once calculated, it will go through the images and 
# correct the UTC times to the correct value.  This process
# assumes that the UTC time for an image right after the 
# second turns is a fiducial time.  In addition, the program
# will fix the timing for a series of frames where the timing
# of the frames is based on the reading out the wrong number
# of frames.
#
# 20080119  -Fixed bug that wrote negative times
# 20080217  -Added the addition fixed to the timing
#

import os,sys, time, re
from pyraf import iraf 
import saltprint, saltkey, saltio, salttime
import slottool
import saltplot

import pyfits
from numpy import *

# ----------------------------------------------------------
# core routine

def slotutcfix(images, update, outfile, ampperccd, ignorexp, droplimit, inter, plotdata, logfile,verbose,status):



#set up the variables
    status = 0
    utc_list = []

# test the logfile

    logfile = saltio.logname(logfile)

# log the call 

    saltprint.line(logfile,verbose)
    history = 'SLOTUTCFIX -- '
    history += 'images='+images+' '
    yn = 'n'
    if (update): yn = 'y'
    history1 = 'update='+yn+' '
    yn = 'n'
    history1 += 'outfile=%s ' % outfile
    history1 += 'ampperccd=%i ' % ampperccd
    history1+= 'ignorexp=%i ' % ignorexp
    history1+= 'droplimit=%i ' % droplimit

    if (inter): yn = 'y'
    history1+= 'inter='+yn+' '
    yn = 'n'
    if (plotdata): yn = 'y'
    history1+= 'plotdata='+yn+' '
    history1+= 'logfile='+logfile+' '
    yn = 'n'
    if (verbose): yn = 'y'
    history1+= 'verbose='+yn
    message= history + history1+'\n'
    saltprint.log(logfile,message,verbose)

# start time

    saltprint.time('SLOTUTCFIX started at  ',logfile,verbose)

# is the input file specified?
   
    if (status == 0): status = saltio.filedefined('Input',images,logfile)

# if the input file is a list, does it exist?

    if (status == 0 and images[0] == '@'): 
        status = saltio.listexists('Input',images,logfile)

# parse list of input files and place them in order

    if (status == 0):
        infiles, status = saltio.listparse('Raw image',images,'','','',logfile)
        infiles.sort()

# check input files exist

    if (status == 0): status = saltio.filesexist(infiles,'','r',logfile)

# check to see if the output file exists and if so, clobber it

    if (status ==0):
       if  os.path.isfile(outfile):
         try: 
          os.remove(outfile)
         except:
          message = 'SLOTUTCFIX--ERROR: File ' + outfile + ' can not be removed'
          status = saltprint.err(logfile,message)

# open the outfile

    if (status == 0 and outfile):
       try:
          fout=open(outfile,'w')
       except:
          message = 'SLOTUTCFIX--ERROR: File ' + outfile + ' can not be opened'
          status = saltprint.err(logfile,message)


# get time of first exposure and basic information about the observations

    if (status == 0):

	infile=infiles[0]
	struct, status = saltio.openfits(infile,logfile)

        # check to make sure slotmode data
        if (status ==0 ):
	   detmode, status = saltkey.get('DETMODE',struct[0], infile, logfile)
	   if detmode != 'Slot Mode':
              message='ERROR--SLOTUTCFIX: Data are not Slot Mode Observations'
              status=saltprint.err(logfile, message)
 
        # Check to see if SLOTUTCFIX has already been run
        if (status == 0 ):
           if saltkey.found('SLOTUTC', struct[0]):
              message='WARNING--SLOTUTCFIX: Data have already been processed by SLOTUTCFIX'
              saltprint.log(logfile, message, verbose)

        #check to make sure that it is the right version of the software
        if status == 0:
           scamver, status = saltkey.get('DETSWV', struct[0], infile, logfile)
           try:
              scamver=float(scamver.split('-')[-1])
              if 4.42 <= scamver <= 5.00:
                  pass
              else:   
                  message='ERROR--SLOTUTCFIX: SLOTUTCFIX cannot currently correct this version of the SCAM software.'
                  status=saltprint.err(logfile, message)
               
           except:
              message='WARNING--SLOTUTCFIX: Not able to read software version'
              saltprint.log(logfile, message, verbose)
        

        #requested exposure time
        if (status == 0):
	   req_texp,status = saltkey.get('EXPTIME',struct[0],infile,logfile)
              

	# how many extensions?
    	if (status == 0): 
	   nextend,status = saltkey.get('NEXTEND',struct[0],infile,logfile)

	# how many amplifiers
    	if (status == 0): 
	   amplifiers,status = saltkey.get('NCCDS',struct[0],infile,logfile)
	   amplifiers = int(ampperccd*float(amplifiers))
	   if ampperccd>0:
	      nframes = nextend/amplifiers
	      nstep=amplifiers
	   else:
	      nframes = nextend
	      nstep=1

	#how many total frame and unique times
	if (status==0):
	   ntotal=nextend*len(infiles)
	   nunique=len(infiles)*nframes-ignorexp+1

    #Create arrays necessary for analysis
    if (status==0):
       id_arr=arange(nunique)
       utc_arr=zeros(nunique,dtype=float)

    #Read in each file and make a list of the UTC values
    if status==0:
      j=0
      for infile in infiles:

	if (status == 0 ): struct, status = saltio.openfits(infile,logfile)
        if (status == 0 ):  
           if not len(struct)-1==nextend:
              message='SALTUTCFIX -- ERROR:  %s has a different number of extensions from the first file' % infile
              status=saltprint.err(logfile, message)

	#Skip through the frames and read in the utc
        if (status == 0 ):
           istart=1
           if infile==infiles[0]:
              istart=ignorexp*amplifiers+1
	   for i in range(istart,len(struct), amplifiers):
            if (status == 0):
             try:
	       utc_list.append(saltkey.get('UTC-OBS', struct[i], infile, logfile)[0])
	       utc_arr[j]=slottool.getobstime(struct[i], infile, logfile)
               j += 1
             except Exception, e:
               message='SALTUTCFIX -- ERROR: Unable to create array of UTC times.  Please check the number of extensions in the files'
               status=saltprint.err(logfile, message)
               

        # close FITS file		
	if (status == 0): status = saltio.closefits(struct,logfile)

    #set up the other important arrays 
    if (status == 0): 
        try:
           diff_arr=utc_arr.copy()
           diff_arr[1:]=diff_arr[1:]-utc_arr[:-1]
	   diff_arr[0]=-1
           dsec_arr=utc_arr-utc_arr.astype(int)
	except:
	   message='SLOTUTCFIX--ERROR:  Unable to create timing arrays'
           status=saltprint.err(logfile, message)
           

    #calculate the real exposure time
    if (status == 0):
	real_expt, med_expt, t_start, t_arr, ysum_arr,  status= \
           calculate_realexptime(id_arr, utc_arr, dsec_arr, diff_arr, req_texp, logfile, utc_list)
	med_expt, real_expt, req_texp, t_start


      
    #plot the results
    if (status==0 and plotdata):
       newplot=saltplot.saltplot(t_arr, ysum_arr, 'Time (s)', 'Fit', '', logfile, verbose)
       #newplot.runplotdata()


    #Calculate the corrrect values
    if (status==0):
       i_start = abs(utc_arr-t_start).argmin()
       t_diff=utc_arr*0.0+real_expt
       nd=utc_arr*0.0
       ndrop=0
       for i in range(len(utc_arr)):
          if utc_arr[i] >= t_start:
            t_new=t_start+real_expt*(i-i_start+ndrop)
            t_diff[i]=utc_arr[i]-t_new
            while (t_diff[i]>real_expt and nd[i] < droplimit):
                  nd[i]+= 1
                  t_new=t_start+real_expt*(i-i_start+ndrop+nd[i])
                  t_diff[i]=utc_arr[i]-t_new
            if (nd[i]<droplimit):
               ndrop += nd[i]
          else: 
            t_new=t_start+real_expt*(i-i_start)
            t_diff[i]=utc_arr[i]-t_new
            while (t_diff[i]>real_expt and nd[i] < droplimit):
                  nd[i]+= 1
                  t_new=t_start+real_expt*(i-i_start-nd[i])
                  t_diff[i]=utc_arr[i]-t_new

    #calculate the corrected timestamp by counting 6 record files forward and 
    #8 recored + unrecorded files back--or just 8*t_exp forward.
    #if the object is near the end of the run, then just replace it with 
    #the correct value assuming no dropped exposures.

    if status==0:
       #first make the array of new times
       new_arr=utc_arr-t_diff

       #Next loop through them to find the corrected time
       corr_arr=utc_arr*0.0
    
       for i in range(len(new_arr)):
           if i+6 < len(new_arr)-1:
             corr_arr[i]=new_arr[i+6]-8*real_expt
           else:
             corr_arr[i]=new_arr[i]-2*real_expt

       t_diff=utc_arr-corr_arr
    
    #write out the first results
    if (status == 0): 
       message="Dwell Time=%5.3f Requested Exposure Time=%5.3f Nobs = %i Dropped = %i" % (real_expt, req_texp,  nunique, ndrop)
       saltprint.log(logfile, message, verbose)
       if outfile:
          fout.write('#'+message+'\n')
          fout.write('#%23s %2s %12s %12s %10s %8s %4s \n' % ('File', 'N', 'UTC_old', 'UTC_new', 'UTC_new(s)', 'Diff', 'drop' ))
       
    #Give the user a chance to update the value
    if (status==0 and inter):
       message='Update headers with a dwell time of %5.3f s [y/n]? ' % real_expt
       update, status=saltio.yn_ask(message, logfile) 
       if not update:
          message='Set Dwell Time manually [y/n]? '
          update, status=saltio.yn_ask(message, logfile) 
          if update:
             message='New Dwell Time: ' 
             real_expt, status = saltio.ask(message, logfile)
             try: 
                real_expt=float(real_expt)
             except Exception, e:
                message='ERROR--SALTUTCFIX: Could not set user dwell time because %s' % e
                status=saltprint.err(logfile, message)
                
             
         
               
    #If requested, update the UTC times
    if (status==0 and (update or outfile)):
      j=0
      for infile in infiles:

	struct, status = saltio.openupdatefits(infile,logfile)

	#Skip through the frames and read in the utc
        if (status == 0 ):
           istart=1
           if infile==infiles[0]:
              istart=ignorexp*amplifiers+1
	   for i in range(istart,len(struct), amplifiers):
            for k in range(0,amplifiers):
              if update:
                 struct[i+k], status=updateheaders(struct[i+k],i+k, t_diff[j], real_expt, utc_list[j], infile, logfile, verbose)
            if outfile: 
               utc_new=saltkey.get('UTC-OBS', struct[i+k], infile, logfile)[0]
               utc_new_sec=slottool.getobstime(struct[i], infile, logfile)
               fout.write('%25s %2i %12s %12s %7.3f  %5.4f %4i \n' % (infile, i, utc_list[j], utc_new, utc_new_sec, t_diff[j], nd[j] ))
            j += 1
        #Housekeeping key words
        if (status == 0 and update):
              history = 'SALTUTCFIX -- '
              history += 'images='+infile+' '
              history += history1
              status = saltkey.housekeeping(struct[0], 'SLOTUTC', 'UTC has been corrected', history,infile,logfile)
       

        # update fits file	
        if (status==0 and update):
            status = saltio.updatefits(struct,logfile)
            status = saltio.closefits(struct,logfile)

#close outfile
    if (status==0 and outfile):
            fout.close()


# end time

    if (status == 0):
	saltprint.time('SLOTUTCFIX completed at',logfile,verbose)
    else:
	saltprint.time('SLOTUTCFIX aborted at',logfile,verbose)


def updateheaders( struct, id, tdiff, real_expt, utc, infile, logfile, verbose):
   status=0
 
   #exit if tdiff wasn't updated
   if tdiff == real_expt:
           message='WARNING--SALTUTCFIX: No adequate correction found for frame %i in file %s' % (id, infile)
           saltprint.log(logfile, message, verbose)
           return struct, status


   #calculate the new utc value
   try:
     ntime=salttime.sex2dec(utc)
     ntime=ntime-tdiff/3600.0
     newutc=salttime.dec2sex(ntime)
   except Exception,e:
     message='ERROR--SALTUTCFIX: Could not update UTC in %i header of image %s because %s' % (id, infile, e)
     status=saltprint.err(logfile, message)
     return struct, status
   
   #update the headers
   if utc== saltkey.get('UTC-OBS', struct, infile, logfile)[0]:
       expt_string='%5.4f' % real_expt
       td_string='%5.4f' % tdiff
       if not saltkey.found('DUTC', struct):
           try: 
                  status = saltkey.put('UTC-OBS', newutc, struct, infile, logfile)
                  status = saltkey.put('TIME-OBS', newutc, struct, infile, logfile)
                  status = saltkey.new('DWETIME', expt_string, 'Dwell Time', struct, infile, logfile)
                  status = saltkey.new('DUTC', td_string, 'Change in UTC time', struct, infile, logfile)
           except Exception, e:
                  message='ERROR--SALTUTCFIX: Could not update %i header of image %s because %s' % (id, infile, e)
                  status=saltprint.err(logfile, message)
       else:
           try: 
                  status = saltkey.put('UTC-OBS', newutc, struct, infile, logfile)
                  status = saltkey.put('TIME-OBS', newutc, struct, infile, logfile)
                  status = saltkey.put('DWETIME', real_expt, struct, infile, logfile)
                  status = saltkey.put('DUTC', tdiff, struct, infile, logfile)
           except Exception, e:
                  message='ERROR--SALTUTCFIX: Could not update %i header of image %s because %s' % (id, infile, e)
                  status=saltprint.err(logfile, message)
   else:
       message='ERROR--SALTUTCFIX: Frame missing from list of times'
       status=saltprint.err(logfile, message)

   return struct, status

# -----------------------------------------------------------
# Calculate the actual exposure time

def calculate_realexptime(id_arr, utc_arr, dsec_arr, diff_arr, req_texp, logfile, utc_list):
   """Calculates the real exposure time.  This makes the following assumptions:
	1. That the measurement after the turn of the second is a fiducial
	2. That there is an integer number of frames between each fiducial exposure
        3. We then set up a metric which is Y=sum(i-int(i)) where i=dt/t_exp
        4. Then the minimum of Y is found between the requested exposure time and 
             the median time difference
        5. And the best exposure time is the time at that minimum
                
  
       returns median exposure time and real exposure time
   """
   status=0
   t_exp=0

   #calculate the median time
   try:
        t_wrong=median(diff_arr)
   except:
	message='SLOTUTCFIX--ERROR:  Unable to calculate median time difference'
        status=saltprint.err(logfile, message)

    
   #Compress the arrays to find those closest to the second mark
   if (status==0):
       mask=(dsec_arr<t_wrong)*(diff_arr>0)
       t=compress(mask,utc_arr)
       s=compress(mask,dsec_arr)
       id=compress(mask,id_arr)

   #Now set up the components in the equation
   if (status==0):
       try:
          t_start=t[0]
          dt=t[1:]-t[0] 
       except Exception, e:
	  message='SLOTUTCFIX--ERROR:  Unable to set up necessary arrays because %s' % e
          status=saltprint.err(logfile, message)
	
   #Now find the one that minimizes the equation y=sum(i-int(i))
   if (status==0):
        t_max=t_wrong*1.2
        if t_max < req_texp+0.05:
           t_max=req_texp+0.15
        try:
	  t_arr, ysum_arr=find_real_time(dt, req_texp, t_max)
          t_real=t_arr[ysum_arr.argmin()]
	except Exception, e:
	  message='SLOTUTCFIX--ERROR:  Unable to calculate real dwell time because %s' % e
          status=saltprint.err(logfile, message)

   return t_real, t_wrong, t_start, t_arr, ysum_arr, status


def find_real_time(dt, t_min, t_max):
   """Find the real exposure+dead time
   
      returns float  
   """
   t_e=arange(t_min,t_max,0.0001)
   ysum=t_e*0.0
   for i in range(len(t_e)):
       ysum[i]=ntime_func(dt, t_e[i])
   return t_e, ysum



   
def ntime_func(dt, t_e):
    """Merit function to determine best time
       Weighted for the number of objects in each step
       return float
    """
    y=0
    for j in range(len(dt)):
	i=dt[j]/t_e
	y += abs(i-round(i))
    return y


# -----------------------------------------------------------
# main code 

parfile = iraf.osfn("slottools$slotutcfix.par") 
t = iraf.IrafTaskFactory(taskname="slotutcfix",value=parfile,function=slotutcfix,pkgname='slottools')
